
#if !defined(SAPS_SAPS_h)
#define SAPS_SAPS_h

/* These interfaces are part of the interface specification. */
/* They are canonical, and should be laid out just the way that */
/* they are described here. */
/* This file contains no platform-specific types (HWND et al), although it */
/* may contains compiler and platform detection code to get size32, size64 */
/* and friends right. */

typedef unsigned char byte;
typedef unsigned int size32;
typedef unsigned long long size64;
typedef unsigned int ChannelMask;
typedef float Hz;
typedef float Pixels;
typedef float Radians;

struct Address
{
  size64          Data;
  void           *UserRef;
};

/* If the system is stopped, it means that no generators will be */
/* generating new sound, but tail may still be generated. */
/* If the system is FreeWheeling, then time and beat keep increasing, */
/* but the timeline will not be paid attention to (for tempo changes, etc). */
/* Real-time (live) systems will pretty much always be free-wheeling. The */
/* tempo/time is there to support tempo-matched effects, live synth playing */
/* etc, but even though time progresses, it's not correlated to anything */
/* permanent. */
/* Playing means that the system is playing back a given time track, so */
/* times/positions are correlated to something physical that is intended to */
/* be repeatable. */
/* Recording is like playing, but the system is "armed" such that tracks may */
/* be punched in/out by the user (not really of interest to the plug-in; it */
/* can pretty much treat it as playing). */
enum TransportState
{
  TransportStateStopped,
  TransportStateFreeWheeling,
  TransportStatePlaying,
  TransportStateRecording,
};

struct TimePos
{
  size64          StructSize;

  /* streaming measurement at given sample rate */
  size64          SamplePos;                      //  Samples played since start of song

  /* MIDI song pos */
  size32          BeatPos;                        //  Beat # n
  size32          TickPos;                        //  Tick (clock) within beat; 24 per quarter note beat

  /* SMTPE timecode */
  byte            HourPos;
  byte            MinutePos;
  byte            SecondPos;
  byte            FramePos;
  size32          SubFramePos;                    //  Which subframe within the frame?

  /* Whether the system is stopped, free-wheeling, playing or recording */
  TransportState  Transport;
};

struct TimeConfig
{
  size64          StructSize;

  size32          TicksPerBeat;                   //  MIDI ticks, typically

  size32          FramesPerSecondNumerator;       //  for example 5994
  size32          FramesPerSecondDenominator;     //  for example 100
  size32          SubFrameCount;                  //  how many sub frames within a frame?
};

/* How tempo changes work: The time signature and tempo at the start and end */
/* times of the buffer represent those instantaneous points in time. If there */
/* is a change in between, that will be marked with an event being sent, as */
/* well. This allows the application to do ritardando without excessive event */
/* sending; the Tempo for begin/end will simply slow. */
struct Tempo
{
  size64          StructSize;

  size32          BeatsPerMinuteNumerator;
  size32          BeatsPerMinuteDenominator;
  size32          TimeSignatureNumerator;     //  how many beats per bar
  size32          TimeSignatureDenominator;   //  what is a "beat" anyway? (1/4, 1/8 etc)
};

struct Color
{
  float R;
  float G;
  float B;
  float A;
};

struct VertexColor
{
  byte            BlueByte;
  byte            GreenByte;
  byte            RedByte;
  byte            AlphaByte;
};

#define SAPS_FLAG_PRERELEASE 0x1
#define SAPS_VERSION_MAJOR 2
#define SAPS_VERSION_MINOR 0
#define SAPS_VERSION_FLAGS SAPS_FLAG_PRERELEASE

struct Version
{
  size32          Flags;
  byte            Major;
  byte            Minor;
  byte            Pad0;
  byte            Pad1;
  Version()
    {
      Flags = SAPS_VERSION_FLAGS;
      Major = SAPS_VERSION_MAJOR;
      Minor = SAPS_VERSION_MINOR;
      Pad0 = 0;
      Pad1 = 0;
    }
};

#define SAPS_CURRENT_VERSION Version()

struct EffectInfo
{
  size32          StructSize;
  wchar_t const  *Name;
  wchar_t const  *Category;
  wchar_t const  *Icon;
  size32          AIn;
  size32          AOut;
  size32          MIn;
  size32          MOut;
};

struct PluginInfo
{
  size64          StructSize;
  wchar_t const  *Name;
  wchar_t const  *Copyright;
  wchar_t const  *SupportInfo;
  wchar_t const  *Icon;
  Version         SAPSVersion;          //  The version of SAPS used
  size32          PluginVersion;        //  Newer releases will always have higher numbers here
};

class ItfEffect;
class ItfPlugin;
class ItfWindow;
class ItfElement;


class ItfConfig
{
  public:
    /* Information about current time. */
    virtual void Time(TimeConfig *oConfig, Tempo *oTempo, TimePos *oCurPos) = 0;
    /* The sampling rate of the host */
    virtual Hz   SampleRate() = 0;
    /* Each call to Process() will process exactly this many samples. */
    virtual size32 BlockSize() = 0;

    virtual void *Extension(wchar_t const *name) = 0;
};

enum BusType
{
  /* by convention, odd-numbered constants are "input" busses. */
  BusTypeNull,
  BusTypeTime,
  BusTypeMidiOut,
  BusTypeMidiIn,
  BusTypeAudioOut,
  BusTypeAudioIn,
};

class ItfAddressable
{
  public:
    /* Where do I post messages to if I want to send a message to this target? */
    virtual Address     TargetAddress() = 0;
};

class ItfBus : public ItfAddressable
{
  public:
    /* The name of this bus. */
    virtual wchar_t const *
                        Name() = 0;
    /* Whether the bus is an input or output bus. */
    virtual bool        IsInput() = 0;
    /* Whence you came. */
    virtual ItfEffect  *Effect() = 0;
};

class ItfAudioBus : public ItfBus
{
  public:
    /* The host will request some number of channels, but that might not be supplied. */
    /* The host *must* call RequestChannelCount() on each audio bus before calling Commit(). */
    virtual bool        RequestChannelCount(size32 count) = 0;
    /* How many channels actually exist on this input/output, after RequestChannelCount()? */
    virtual size32      ChannelCount() = 0;
    /* Set up the pointer to use for this particular bus and a particular channel. */
    /* The host *must* call SetPointer() with a proper pointer for each existing channel */
    /* after RequestChannelCount() but before Commit(). */
    virtual void        SetPointer(size32 channel, float *ptr) = 0;
    /* the ChannelMask describes what each channel is (left, right, center, top-right, etc) */
    virtual void        SetChannelMask(ChannelMask mask) = 0;
    /* the ChannelMask describes what each channel is (left, right, center, top-right, etc) */
    virtual ChannelMask BusChannelMask() = 0;
    /* How much delay is introduced by this bus? This is in addition to the fact that the */
    /* bus processes in blocks. A purely linear bus like the "gain" sample introduces zero */
    /* additional delay, and returns 0 here. */
    virtual size32      SamplesDelay() = 0;
};

class ItfMidiBus : public ItfBus
{
  public:
    /* MIDI data is delivered as ItfEvent events, not as data. */
};

/*  The data pointed at by these pointers will be vaild from the call to Commit() */
/* to the next call to Reconfig(). The actual TimeBusConfig structure will not be */
/* valid after the call to SetPointers(), so you have to copy it. */
struct TimeBusConfig
{
  size64          StructSize; //  sizeof(TimeBusConfig)
  TimePos const  *beginTime;  //  time for the first sample in the Process call
  TimePos const  *endTime;    //  time after the last sample in the Process call
  Tempo const    *beginTempo; //  tempo at the beginning of the Process call
  Tempo const    *endTempo;   //  tempo at the end of the Process call
};

class ItfTimeBus : public ItfBus
{
  public:
    /* The host must call SetPointers() on each ItfTimeBus before calling Commit(). */
    /* The host will update the time information before calling Process each time, in */
    /* a matter that is thread-safe with respect to calls into this effect instance. */
    virtual void        SetPointers(TimeBusConfig const &config) = 0;
};


enum EventType
{
  EventTypeNull = 0,        //  no event
  EventTypeMidi = 1,        //  a MIDI event
  EventTypeTempo = 2,       //  tempo
  EventTypePreSeek = 3,     //  advisory about upcoming seek
  EventTypeSeek = 4,        //  actual seek event
  EventTypeAutomation = 5,  //  a float value
  EventTypeText = 6,        //  a text value
  EventTypePlay = 7,        //  means song pos will advance
  EventTypePause = 8,       //  means song pos will not advance
  EventTypePunchIn = 9,     //  means recording starts
  EventTypePunchOut = 10,   //  means recording stops
  EventTypeArm = 11,        //  means the track is being prepared for recording
};

struct EventMidi
{
  size64        StructSize; //  sizeof(MidiEvent)
  unsigned char const *
                LargeData;  //  data if size is > 4 (points at Data if smaller)
  size32        Size;       //  size of data
  unsigned char Data[4];    //  data if size is <= 4, else filled with 00
};

struct EventTempo
{
  size64        StructSize; //  sizeof(TempoEvent)
  size32        DurationSamples;    //  how long does it take to reach target tempo?
  Tempo         TargetTempo;//  target tempo and time signature (time sig changes exactly when durationsamples have elapsed)
};

struct EventPreSeek
{
  size64        StructSize; //  sizaeof(EventPreSeek)
  TimePos       ToTime;     //  to what time does the seek happen
  TimePos       AtTime;     //  at what time does the seek happen
};

struct EventSeek
{
  size64        StructSize; //  sizaeof(EventSeek)
  TimePos       ToTime;     //  to what time does the seek happen
};

struct EventAutomation
{
  size64        StructSize; //  sizeof(EventAutomation)
  size32        DurationSamples;  //  how long until the value is reached?
  float         Value;      //  value for automation
};

struct EventText
{
  size64        StructSize; //  sizeof(EventText)
  wchar_t const *
                LargeData;  //  points at Data, or the actual data
  size32        Length;     //  length of string, not including terminating 0
  wchar_t       Data[54];   //  short strings
};

struct EventPlay
{
  size64        StructSize;
  bool          IsRecording;  //  true if there may be punch in/out events
};

struct EventArm
{
  size64        StructSize;
  bool          WillRecord;   //  true if the track will be recorded, false if not
};


class ItfEvent
{
  public:
    /* Each event happens at some sample offset within the process call. */
    virtual size32      Offset() = 0;
    virtual EventType   Type() = 0;
    virtual void       *EventData() = 0;
    virtual size64      EventSize() = 0;

    /* This pointer is reserved for use by the event recipient. The support library */
    /* will stuff a pointer used for dispatching here. The sender should send this as 0. */
    void *UserData;
};

enum ParameterType
{
  ParameterTypeNull = 0,
  ParameterTypeFloat = 1,
  ParameterTypeInt = 2,
  ParameterTypeText = 3,
  ParameterTypeBinary = 4,
};

class ItfParameter : public ItfAddressable
{
  public:
    /* What type of parameter? Float, Int, etc. */
    virtual ParameterType   Type() = 0;
    /* the actual value of the parameter */
    virtual void const     *Value() = 0;
    /* what is the size of the parameter value? This is in bytes, and includes the terminating 0 for strings. */
    virtual size64          Size() = 0;
    /* Update the value of the parameter (pointing at a float, size64, wchar_t string or a block of memory) */
    virtual void            SetValue(void const *value, size64 size) = 0;
};

class ItfBlock
{
  public:
    /* Read the block -- a pointer to the beginning of the data. */
    virtual void const *Data() = 0;
    /* The size of the data. */
    virtual size64 Size() = 0;
    /* When the owner is done, he/she calls Dispose() which disposes or recycles the block. */
    virtual void Dispose() = 0;
};

class ItfWriteBlock
{
  public:
    /* To create an ItfBlock that you can write to, create it on the ItfHost, and then */
    /* call Write() to receive a pointer into the block of at least a given size. You can */
    /* optionally make that size live somewhere other than the start. The total size of */
    /* the block will be set to size + offset, if the size is currently smaller. */
    /* If the block was pre-allocated with a certain physical size other than 0, attempting to extend */
    /* the block past the physical size will fail, and return NULL. */
    virtual void *Write(size64 size, size64 offset = 0) = 0;
    /* Retrieve the actual block that this writable block represents. That's the only way you can dispose it. */
    virtual ItfBlock *Block() = 0;
};

class ItfEventTarget
{
  public:
    /* Things that can be address targets register themselves as ItfEventTarget interfaces. */
    /* In general, this is done by the helper classes for bus, effect, GUI etc. */
    /* When the events received are real-time, they are guaranteed good until the end of */
    /* the next Process() call sequence. When the events are not real-time, they are guaranteed good */
    /* until the end of the next Idle() call sequence. */
    virtual void    DoEvents(size32 count, ItfEvent **events) = 0;
};

/* Note that the Effect is not addressable, but the busses and parameters are! */
class ItfEffect : public ItfEventTarget
{
  public:
    /* The host is wanting to configure you. Please set yourself up, if you haven't already. */
    /* The host cal call Bus() after Reconfig() and before Commit(). */
    /* The host guarantees that the plug-in is not in real-time mode while doing config. */
    virtual void Reconfig(ItfConfig *config) = 0;
    /* How many busses does the Effect have? Each bus can be audio or midi (or something else), and input or output. */
    virtual size32 BusCount() = 0;
    /* Return the interface for a given bus, or NULL for a bad index. */
    virtual ItfBus *Bus(size32 index) = 0;
    /* The host is done re-configuring you -- after this, Bus() won't be called until after another call to Reconfig(). */
    virtual void Commit() = 0;

    /* The host calls Idle() every so often, even while the effect is in real-time mode. */
    /* The estimated frequency is how many milliseconds between each call to Idle the */
    /* effect can expect on average. */
    virtual void Idle(size32 estMs) = 0;
    /* An effect is requested to open a GUI. It should turn around and call OpenWindow(), or just return NULL. */
    /* It is NOT OK for an effect to implement ItfWindow itself, it must use a window as it comes from the host. */
    virtual ItfWindow *OpenGui() = 0;
    /* How much tail is there on this effect? */
    virtual size32 SamplesTail() = 0;

    /* In the real-time thread, before any processing is done, EnterRealtime() is called. */
    /* The host guarantees to not reconfig the plug-in while in real-time mode. */
    /* You can also use EnterRealtime() as a "pre-roll" call to make sure everything is "spun up." */
    virtual void EnterRealtime(TimePos const &atPos) = 0;
    /* Process() is called at a regular interval to generate or process audio and MIDI. */
    /* Tempo events are provided as an Events. */
    virtual void Process() = 0;
    /* LeaveRealtime() is called last in the real-time thread, before processing in the host is suspended. */
    virtual void LeaveRealtime() = 0;

    /* When the host is done with the effect, it calls Dispose(). This never happens in real-time mode. */
    virtual void Dispose() = 0;
    /* Whence did you come? */
    virtual ItfPlugin *Plugin() = 0;

    virtual void *Extension(wchar_t const *name) = 0;
};

/* ItfPlugin is the interface that a plug-in (DLL) provides to the host application */
class ItfPlugin
{
  public:
    /* The host is done with the plug-in, and may unload it. */
    virtual void              Unload() = 0;
    /* The plug-in returns some information about itself. */
    virtual PluginInfo const *Info() = 0;
    /* The plug-in can create some number of effects with some parameters. */
    virtual size32            CountEffects() = 0;
    /* For each effect, describe its name, icon, and capabilities. */
    virtual EffectInfo const *Info(size32 index) = 0;
    /* Create an instance of a given effect. */
    virtual ItfEffect        *MakeEffect(size32 index) = 0;

    virtual void             *Extension(wchar_t const *name) = 0;
};

/* Note that the slave is not addressable, but the window is! */
class ItfWindowSlave : public ItfEventTarget
{
  public:
    /* The host will call Opened as the first call into the ItfWindowSlave. This will be before */
    /* the call to OpenWindow has returned. The ItfWindow pointer is the pointer that will */
    /* be returned from OpenWindow(). */
    virtual void    Opened(ItfWindow *window) = 0;
    /* Whenever the mouse enters the window after having been out, MouseEnter is called. */
    virtual void    MouseEnter() = 0;
    /* Whenever the mouse is over a different element than before, MouseOverElement() is called. */
    virtual void    MouseOverElement(ItfElement *elem) = 0;
    /* Whenever the mouse moves at least one pixel (but maybe more), MosueMove() is called. /*
          /* Coordinates are relative to window top-left. */
    virtual void    MouseMove(size32 x, size32 y, size32 buttons) = 0;
    /* Whenever the mouse button (any button) is pressed, MouseDown() is called. The position is */
    /* as per the previous call to MouseOverElement() and MouseMove(). */
    virtual void    MouseDown(size32 oldButtson, size32 newButtons) = 0;
    /* Whenever the mouse is released, MouseUp() is called with the old and new button mask. */
    virtual void    MouseUp(size32 oldButtson, size32 newButtons) = 0;
    /* Whenever the mouse moves out of the window (or the window becomes inactive), MouseLeave() is called. */
    virtual void    MouseLeave() = 0;
    /* If the window has capture, and loses it, CaptureLost() is called. */
    virtual void    CaptureLost() = 0;
    /* If the window is resizable, when the user starts resizing, BeginResize() will be called. */
    virtual void    BeginResize() = 0;
    /* If the window is resizable, while the user is resizing the window, UpdateResize() is called. */
    virtual void    UpdateResize() = 0;
    /* If the window is resizable, after the user has completed resizing, CommitResize() is called. */
    virtual void    CommitResize() = 0;
    /* If the window has focus, and loses it, FocusLost() is called. */
    virtual void    FocusLost() = 0;
    /* Whenever the window has focus and a key is pressed, KeyDown() is called. This does not include printable keys (as will be delivered through Type). */
    virtual void    KeyDown(size32 key, size32 buttons) = 0;
    /* Whenever the window has focus and a key is released, KeyUp() is called. */
    virtual void    KeyUp(size32 key, size32 buttons) = 0;
    /* Whenever the window has focus and text is typed, Type() is called. This does not include non-printable keys (like F-keys etc). */
    virtual void    Type(wchar_t const *text) = 0;
    /* Every once in a while, the window Idle() function will be called. */
    virtual void    Idle() = 0;
    /* When the window is finally closed. Closed() will be called. Typically inside a call to Close(), or as the result of a user click in the close button. */
    /* When Closed() has returned, the host will no longer call this ItfWindowSlave. */
    virtual void    Closed() = 0;

    virtual void   *Extension(wchar_t const *name) = 0;
};

enum TextAdjust
{
  TextAdjustNormal = 0,
  TextAdjustCenter = 1,
  TextAdjustReverse = 2,
};

struct TextInfo
{
  size64            StructSize;   //  sizeof(TextInfo)
  wchar_t const    *Text;         //  non-terminated string
  size32            Length;       //  wcslen(text)
  Pixels            MaxWidth;     //  Maximum width. If 0, use window width.
  Pixels            MaxHeight;    //  Maximum height. If 0, use infinite height.
  Pixels            PointsSize;   //  How many points is the text? (9pt is small but readable; 12 pt is common)
  bool              Bold;         //  Is the text fattened?
  bool              Italic;       //  Is the text italic/slanted/oblique?
  bool              Monospace;    //  Is the text monospace?
  bool              FullRect;     //  Will the element always use the full rect specified by MaxWidth/MaxHeight, 
                                  //  even if it fits in a smaller rect?
  bool              RightToLeft;  //  Is the text in a right-to-left script?
  bool              TopToBottom;  //  Is the text in a top-to-bottom script?
  byte              Adjust;       //  Normal (Left for Latin), Center, Reverse (Right for Latin)
};

class ItfElement
{
  public:
    /* Remove the element, and dispose it so it can't be used again. */
    virtual void    Dispose() = 0;
    /* Return the window the element is attached to, or NULL. */
    virtual ItfWindow *
                    Window() = 0;
    /* Remove the element from its window (if any), but keep it around (you can later add it back). */
    virtual void    Remove() = 0;
    /* Given a window coordinate, test whether the element was hit by that point, and if so, */
    /* calculate element-relative coordinates (which is not trivial for rotated elements). */
    virtual bool    PointHit(Pixels x, Pixels y, Pixels *ox, Pixels *oy) = 0;
    /* Return the reference point, in element coordinates, around which the element is rotated and positioned. */
    virtual void    RefPoint(Pixels *ocx, Pixels *ocy) = 0;
    /* Set the reference point, in element coordinates, around which the element is rotated, and which is positioned with SetPosition(). */
    virtual void    SetRefPoint(Pixels cx, Pixels cy) = 0;
    /* Return the position, in window coordinates, of the reference point of the element. */
    virtual void    Position(Pixels *ox, Pixels *oy) = 0;
    /* Set the position, in window coordinates, of the reference point of the element. */
    virtual void    SetPosition(Pixels x, Pixels y) = 0;
    /* Return the size, in element coordinates, of the element. */
    virtual void    Extent(Pixels *ow, Pixels *oh) = 0;
    /* Return the rotation, in radians counter-clockwise, of the element. */
    virtual Radians Rotation() = 0;
    /* Set the rotation, in radians counter-clockwise, of the element. */
    virtual void    SetRotation(Radians r) = 0;
    /* Return the scale, horizontal and vertical in widget coordinate space (scale axis rotates with widget). */
    virtual void    Scale(float *osx, float *osy) = 0;
    /* Set the scale, horizontal and vertical in widget coordinate space (scale axis rotates with widget). */
    virtual void    SetScale(float sx, float sy) = 0;
    /* Return the tint of the element. The "add" value is added to each pixel, the "multiply" value is multiplied into each pixel. */
    /* Note that these values may be greater than 1, or less than 0! */
    virtual void    Tint(Color *outAdd, Color *outMultiply) = 0;
    /* Set the tint. The "add" values are added to each pixel; the "multiply" value is multiplied into each pixel. */
    virtual void    SetTint(Color add, Color multiply) = 0;
    /* Turn off or on whether the slave should receive notifications about the mouse moving over this element. */
    virtual void    SetWantsMouse(bool wantsMouse) = 0;
    /* Return whether the slave should receive notifications about the mouse moving over this element. */
    virtual bool    WantsMouse() = 0;

    /*  Aah! Some data in an interface! This data is for use by the effect/plug-in implementer, and not touched by the library or host. */
    void           *UserPtr;
};

struct BitmapInfo
{
  void const       *Pixels;     //  Pointer to pixels. Pixels are stored in the BGRA byte order.
  size32            Pitch;      //  how many bytes per row (which may be more than 4 * width)
  size32            Width;      //  how many columns of pixels. Column 0 is at the left.
  size32            Height;     //  how many rows of data. Row 0 is at the top.
};

class ItfBitmap
{
  public:
    /* Return the element that can actually be added/positioned. */
    virtual ItfElement *
                    Element() = 0;
    /* Fill the bitmap, or a part of it, with some data. */
    virtual void    SetData(BitmapInfo const &info, size32 left = 0, size32 top = 0) = 0;
};

class ItfText
{
  public:
    /* Return the Element that can actually be placed in the window. */
    virtual ItfElement *
                    Element() = 0;
    /* Update the text of the element. */
    virtual void    SetText(TextInfo const &info) = 0;
    /* Return the text */
    virtual wchar_t const *
                    Text() = 0;
    /* Given a point (in widget-relative coordinates), return which character that points at. */
    /* This is offset half a character, to be properly used when drag-selecting text. */
    virtual size32  CharacterFromPoint(Pixels x, Pixels y) = 0;
    /* How tall is a line in the current text? */
    virtual Pixels  LineHeight() = 0;
    /* Given a character index (from 0 to length-1), what is the rectangle that encloses that character? */
    virtual void    RectFromCharacter(size32 ix, Pixels *oLeft, Pixels *oRight, Pixels *oTop, Pixels *oBottom) = 0;
    /* Given a character index (from 0 to length-1), what line is that character part of? */
    virtual size32  LineFromCharacter(size32 ix) = 0;
    virtual size32  LineCount() = 0;
    /* Given a line, what index does the character have that starts that line? */
    virtual size32  LineStartCharacter(size32 line) = 0;
    /* Get the selection (highlighted area). If there is no selection, return false. */
    virtual bool    Selection(size32 *ostart, size32 *oend) = 0;
    /* Set the selection (highlighted area). */
    virtual void    SetSelection(size32 start, size32 end) = 0;
};

struct Vertex
{
  Pixels        X;
  Pixels        Y;
  VertexColor   C;
};

enum VertexListMode
{
  VertexListModeNone,             //  not valid
  VertexListModePoints,           //  1 vert per point
  VertexListModeLineList,         //  2 verts per line
  VertexListModeConvexPolygon,    //  1 vertex per corner of the polygon
};

/* A vertex list is a primitive, such as a line list or a polygon, that is rendered in color by the host. */
class ItfVertexList
{
  public:
    /* return the Element that you can position, rotate and scale */
    virtual ItfElement *Element() = 0;
    /* Set the data of the vertex list */
    virtual void SetVertices(Vertex const *v, size32 count) = 0;
    /* how many vertices in this list? */
    virtual size32 VertexCount() = 0;
    /* retrieve the vertex data */
    virtual void Vertices(Vertex *ov, size32 count) = 0;
    /* test which mode the vertex list is in */
    virtual VertexListMode Mode() = 0;
    /* set a new mode for the vertex list */
    virtual void SetMode(VertexListMode mode) = 0;
};


/* Messages that go to the ItfWindow are dispatched to the ItfWindowSlave. */
class ItfWindow : public ItfAddressable
{
  public:
    /* Request a close of the window. */
    virtual void    Close() = 0;
    /* Find the slave of this window. */
    virtual ItfWindowSlave *
                    Slave() = 0;
    /* Resize the window to a given size. The size will be clipped to the min/max size specified on creation. */
    virtual void    Resize(size32 width, size32 height) = 0;
    /* Find out how big the window is now. */
    virtual void    Extent(size32 *ox, size32 *oy) = 0;
    /* Set the mouse capture. This means that all mouse move events will go to this window, even if outside the window bounds! */
    virtual void    SetCapture() = 0;
    /* Set the focus. This means that further typing will go to this window. */
    virtual void    SetFocus() = 0;
    /* Make a bitmap widget, which you can then fill with data and add to the window. */
    virtual ItfBitmap *
                    MakeBitmap(size32 width, size32 height) = 0;
    /* Make a text widget, which you can then add to the window, as well as re-set text on. */
    virtual ItfText *
                    MakeText(TextInfo const &info) = 0;
    /* Make a vertex list widget */
    virtual ItfVertexList *
                    MakeVertexList() = 0;
    /* Add a given element to the window. */
    virtual void    AddElement(ItfElement *widget, ItfElement *before = 0) = 0;
    /* How many elements are in the window? */
    virtual size32  ElementCount() = 0;
    /* Return the element at a given index, or NULL. */
    virtual ItfElement *
                    Element(size32 index) = 0;
    /* Invalidate some part of the window, or the entire window */
    virtual void    Invalidate(ItfElement *elem = 0) = 0;

    virtual void *Extension(wchar_t const *name) = 0;
};

/* Configuration for a window, provided to the host when creating the window */
struct WindowConfig
{
  size64            StructSize;   //  sizeof(WindowConfig)
  wchar_t const    *Title;        //  The title of the window (0-terminated string)
  size32            MinWidth;     //  Minimum width of the window
  size32            MinHeight;    //  Minimum height of the window
  size32            MaxWidth;     //  Minimum width of the window. If <= MinWidth, not resizable horizontally
  size32            MaxHeight;    //  Maximum height of the widnow. If <= MinHeight, not resizable vertically
};

/* The interface that a host provides to a plug-in/effect set */
class ItfHost : public ItfAddressable
{
  public:
    /* Allocate a block for the host. The "safe" size is writable without further memory allocations. */
    /* If you pass 0, then the block is infinitely extendable, but doing so will cause memory allocation, */
    /* and thus should not be done during real-time processing. In general, NewBlock() should not be */
    /* called from the real-time thread. */
    virtual ItfWriteBlock *NewBlock(size64 safeSize) = 0;
    /* The host can create a new window that hosts this effect. That window may be a top-level frame */
    /* window, or it may be a sub-window within a bigger host conglomeration. The "slave" will receive window events. */
    /* In general, OpenWindow() should not be called from the real-time thread. */
    virtual ItfWindow *OpenWindow(WindowConfig const &spec, ItfWindowSlave *slave) = 0;
    /* Send an event to some address. This is a real-time message if atTime is not NULL. */
    virtual bool SendData(Address const &addr, size32 type, ItfBlock *block, TimePos const *atTime = 0) = 0;
    /* Convenience function to not have to allocate an ItfBlock. This is a real-time message if atTime is not NULL. */
    virtual bool SendData(Address const &addr, size32 type, void const *data, size64 size, TimePos const *atTime = 0) = 0;
    /* Allocate an address. This should not be called from the real-time thread. */
    virtual void AllocateAddress(Address *oAddr, void *userRef, ItfEventTarget *target) = 0;
    /* Deallocate an address and clear out the argument. This should not be called from the real-time thread.  */
    virtual void DeallocateAddress(Address &addr) = 0;

    virtual void *Extension(wchar_t const *name) = 0;
};

extern "C" {
/* Every plug-in defines this function. Note that, because it's never defined */
/* in a host, declaring it dllexport always is safe, and makes building plug-ins */
/* less error-prone. */
__declspec(dllexport) ItfPlugin *SAPSCreatePlugin(Version const &ver, ItfHost *host);
}

#endif    //  SAPS_SAPS_h

