
#if !defined(SAPS_SupportLib_h)
#define SAPS_SupportLib_h

/* The supportlib must be compiled with the same version of MSVC as */
/* you compile your plug-in. This is to make sure that class sizes */
/* are the same, especially for STL containers. */

#include "SAPS.h"

#include <vector>
#include <list>
#include <string>

//  This is not built as a DLL, because that would lead to DLL Hell for 
//  different plug-ins built with different DLL versions. Either you'd 
//  have to install them SxS (which is a pain), or build each version of 
//  the library, for each compiler, with a different name (which is a pain).
#define SAPS_SUPPORTLIB_API 

/* Classes used in the implementation of a plug-in, but not strictly required */
/* from the point of view of the specification. */

class PluginBase;

/* convenience helper to clear out a struct and set up its StructSize field. */
template<typename T> inline void Zero(T &t) { memset(&t, 0, sizeof(t)); t.StructSize = sizeof(t); }

/* EventChain is an implementation base class that allows event handlers in the plug-in to delegate. */
class EventChain
{
  public:
    EventChain(EventChain *next = 0)
      : next_(next)
      {
      }
    /* Override this to handle events you can handle, then call base to dispatch to the next for those you can't. */
    virtual void DoEvent(ItfEvent *ev)
      {
        if (next_ != 0)
          next_->DoEvent(ev);
      }
    /* Configure the next handler in the chain. */
    void SetNext(EventChain *next)
      {
        next_ = next;
      }
  private:
    EventChain *next_;
};

/* MidiParser looks at MIDI events and dispatches them to a MidiListener. */
/* Non-MIDI events are sent down the EventChain. */
class MidiParser : public EventChain
{
};

class EffectBase;

/* BusConfig is the base class for all bus information (MidiInput, AudioOutput, etc) */
/* You actually use subclasses such as AudioInput to properly initialize it. */
struct BusConfig
{
  protected:
    inline BusConfig(size64 size, BusType type)
      : StructSize(size)
      , Type(type)
      {
      }
  public:
    size64          StructSize;
    BusType         Type;
};

/* middle class in the type hierarchy of bus configurations */
struct AudioBusConfig : public BusConfig
{
  protected:
    inline AudioBusConfig(BusType type, size32 minCount, size32 maxCount, size32 latency = 0)
      : BusConfig(sizeof(AudioBusConfig), type)
      , MinChannelCount(minCount)
      , MaxChannelCount(maxCount > minCount ? maxCount : minCount)
      , InputLatencySamples(latency)
      {
      }
  public:
    size32        MinChannelCount;
    size32        MaxChannelCount;
    size32        InputLatencySamples;
};

/* Define an audio input bus, with a given minimum and maximum allowable channel count. */
struct AudioInput : AudioBusConfig
{
  inline AudioInput(size32 lo = 2, size32 hi = 0)
    : AudioBusConfig(BusTypeAudioIn, lo, hi)
    {
    }
};

/* Define an audio output bus, with a given minimum and maximum allowable channel count. */
struct AudioOutput : AudioBusConfig
{
  inline AudioOutput(size32 lo = 2, size32 hi = 0)
    : AudioBusConfig(BusTypeAudioOut, lo, hi)
    {
    }
};

/* Define a midi input bus. */
struct MidiInput : BusConfig
{
  inline MidiInput()
    : BusConfig(sizeof(MidiInput), BusTypeMidiIn)
    {
    }
};

/* Define a midi output bus. */
struct MidiOutput : BusConfig
{
  inline MidiOutput()
    : BusConfig(sizeof(MidiInput), BusTypeMidiOut)
    {
    }
};

/* Define a time info input bus. */
struct TimeInput : BusConfig
{
  inline TimeInput()
    : BusConfig(sizeof(TimeInput), BusTypeTime)
    {
    }
};


class BusImpl
{
  public:
    SAPS_SUPPORTLIB_API                         BusImpl(EffectBase *effect, void *userRef);
    SAPS_SUPPORTLIB_API                         ~BusImpl();

    virtual ItfBus                             *Bus() = 0;
    virtual void                                Reconfig(ItfConfig *cfg) = 0;

    Address                                     address_;
    EffectBase                                 *effect_;
    void                                       *userRef_;
};


/* AudioBus is an implementation of the ItfAudioBus interface. These are allocated */
/* by the AddBus() function when adding an audio type bus. */
class AudioBus : public ItfAudioBus, public BusImpl
{
  public:
    SAPS_SUPPORTLIB_API                         AudioBus(AudioBusConfig const &config, wchar_t const *name, EffectBase *base, void *userRef = 0);
    SAPS_SUPPORTLIB_API                         ~AudioBus();

    /* AudioBus interface */
    SAPS_SUPPORTLIB_API float                  *Channel(size32 index);

    /* BusImpl interface */
    SAPS_SUPPORTLIB_API ItfBus                 *Bus();
    SAPS_SUPPORTLIB_API void                    Reconfig(ItfConfig *cfg);

    /* ItfBus interface */
    SAPS_SUPPORTLIB_API Address                 TargetAddress();
    SAPS_SUPPORTLIB_API wchar_t const          *Name();
    SAPS_SUPPORTLIB_API bool                    IsInput();
    SAPS_SUPPORTLIB_API ItfEffect              *Effect();

    /* ItfAudioBus interface */
    SAPS_SUPPORTLIB_API bool                    RequestChannelCount(size32 count);
    SAPS_SUPPORTLIB_API size32                  ChannelCount();
    SAPS_SUPPORTLIB_API void                    SetPointer(size32 channel, float *ptr);
    SAPS_SUPPORTLIB_API void                    SetChannelMask(ChannelMask mask);
    SAPS_SUPPORTLIB_API ChannelMask             BusChannelMask();
    SAPS_SUPPORTLIB_API size32                  SamplesDelay();

    AudioBusConfig                              config_;
    std::wstring                                name_;

    float                                     **pointers_;
    size32                                      channelCount_;
    size32                                      channelMask_;
    float                                      *dummyBuffer_;   //  used when the host is bad
};

/* MidiBus is an implementation of the ItfMidiBus interface. These are allocated */
/* by the AddBus() function when adding a MIDI type bus. */
class MidiBus : public ItfMidiBus, public BusImpl
{
  public:
    SAPS_SUPPORTLIB_API                         MidiBus(BusConfig const &config, wchar_t const *name, EffectBase *base, void *userRef);
    SAPS_SUPPORTLIB_API                         ~MidiBus();

    /* BusImpl interface */
    SAPS_SUPPORTLIB_API ItfBus                 *Bus();
    SAPS_SUPPORTLIB_API void                    Reconfig(ItfConfig *cfg);

    /* ItfBus interface */
    SAPS_SUPPORTLIB_API Address                 TargetAddress();
    SAPS_SUPPORTLIB_API wchar_t const          *Name();
    SAPS_SUPPORTLIB_API bool                    IsInput();
    SAPS_SUPPORTLIB_API ItfEffect              *Effect();

    BusConfig                                   config_;
    std::wstring                                name_;
};

/* TimeBus is an implementation of the ItfTimeBus interface. These are allocated */
/* by the AddBus() function when adding a time type bus. */
class TimeBus : public ItfTimeBus, public BusImpl
{
  public:
    SAPS_SUPPORTLIB_API                         TimeBus(BusConfig const &config, wchar_t const *name, EffectBase *base, void *userRef);
    SAPS_SUPPORTLIB_API                         ~TimeBus();

    SAPS_SUPPORTLIB_API ItfBus                 *Bus();
    SAPS_SUPPORTLIB_API void                    Reconfig(ItfConfig *cfg);

    /* ItfBus interface */
    SAPS_SUPPORTLIB_API Address                 TargetAddress();
    SAPS_SUPPORTLIB_API wchar_t const          *Name();
    SAPS_SUPPORTLIB_API bool                    IsInput();
    SAPS_SUPPORTLIB_API ItfEffect              *Effect();

    /* ItfTimeBus interface */
    SAPS_SUPPORTLIB_API void                    SetPointers(TimeBusConfig const &config);

    BusConfig                                   config_;
    std::wstring                                name_;
    TimeBusConfig                               pointers_;
};

/* Configure how values are turned into text and vice versa */
struct FloatResponse
{
    size64      StructSize;
    /* Text is used to display the value. */
    bool        (*TextToValue)(FloatResponse const *resp, wchar_t const *text, float *oValue);
    bool        (*ValueToText)(FloatResponse const *resp, float iValue, wchar_t *oText, size32 count);
    /* A GUI value always goes from 0 .. 1 linearly */
    float       (*WarpFromGuiToValue)(FloatResponse const *resp, float guiValue, float minValue, float maxValue);
    float       (*WarpFromValueToGui)(FloatResponse const *resp, float value, float minValue, float maxValue);
};

/* A pre-canned response that just uses the value as-is */
SAPS_SUPPORTLIB_API FloatResponse const &LinearResponse();
/* A pre-canned response that translates between dB(FS) and multipliers */
SAPS_SUPPORTLIB_API FloatResponse const &DbResponse();


/* A parameter that is represented as a float */
struct FloatRange
{
  FloatRange(float from, float to, FloatResponse const &resp)
    : From(from)
    , To(to)
    , Response(resp)
    {
      StructSize = sizeof(*this);
    }

  size64            StructSize;
  float             From;
  float             To;
  FloatResponse const &
                    Response;
};

/* Configure how values are turned into text and vice versa */
struct IntResponse
{
    size64      StructSize;
    /* Text is used to display the value. */
    bool        (*TextToValue)(IntResponse const *resp, wchar_t const *text, size64 *oValue);
    bool        (*ValueToText)(IntResponse const *resp, size64 iValue, wchar_t *oText, size32 count);
    void       *ResponseData;
    size32      ResponseCount;
};

/* A parameter that is represented as an size64 */
struct IntRange
{
  IntRange(size64 from, size64 to, IntResponse const &resp)
    : From(from)
    , To(to)
    , Response(resp)
    {
      StructSize = sizeof(*this);
    }

  size64            StructSize;
  size64            From;
  size64            To;
  IntResponse const &
                    Response;
};


/* Given a set of options that you can choose between, initialize a given IntResponse structure */
/* with the necessary functions. Note that the Options array cannot change or move while this */
/* IntResponse is active. */
SAPS_SUPPORTLIB_API IntResponse &SelectorIntResponse(IntResponse *iResponse, wchar_t const *const *Options, size32 Count);
/* Get a On/Off response */
SAPS_SUPPORTLIB_API IntResponse &OnOffIntResponse();
/* Get a response that just uses the integer value. */
SAPS_SUPPORTLIB_API IntResponse &LinearIntResponse();


class ParameterImpl
{
  public:
    SAPS_SUPPORTLIB_API                         ParameterImpl(EffectBase *effect, void *userRef);
    SAPS_SUPPORTLIB_API                         ~ParameterImpl();

    Address                                     address_;
    EffectBase                                 *effect_;
    void                                       *userRef_;
};

/* A FloatParameter represents some quantity that can be varied continuously. */
class FloatParameter : public ItfParameter, public ParameterImpl
{
  public:
    SAPS_SUPPORTLIB_API                         FloatParameter(FloatRange const &fr, wchar_t const *name, wchar_t const *unit, EffectBase *base, void *userRef);
    SAPS_SUPPORTLIB_API                         ~FloatParameter();

    /* ItfParameter */
    SAPS_SUPPORTLIB_API Address                 TargetAddress();
    SAPS_SUPPORTLIB_API ParameterType           Type();
    SAPS_SUPPORTLIB_API void const             *Value();
    SAPS_SUPPORTLIB_API size64                  Size();
    SAPS_SUPPORTLIB_API void                    SetValue(void const *value, size64 size);

    SAPS_SUPPORTLIB_API float                   Read(size32 atOffset);

    FloatRange const                           &range_;
    std::wstring                                name_;
    std::wstring                                unit_;
    float                                       value_;
};

/* An IntParameter represents some quantity that can be varied discretely. */
/* This may be a selector, or a toggle, for example. */
class IntParameter : public ItfParameter, public ParameterImpl
{
};

/* A TextParameter contains some amount of text -- often a filename, say. */
class TextParameter : public ItfParameter, public ParameterImpl
{
};

/* A BinaryParameter contains information that only makes sense to the plug-in, */
/* not to the host. Binary parameters cannot contain references outside itself, */
/* because they may be saved/restored by the host. */
class BinaryParameter : public ItfParameter, public ParameterImpl
{
};

struct EffectInfoBase : public EffectInfo
{
  public:
    EffectInfoBase(wchar_t const *name, wchar_t const *category)
      {
        StructSize = sizeof(EffectInfo);
        Name = (name);
        Category = (category);
        Icon = (0);
        AIn = (0);
        AOut = (0);
        MIn = (0);
        MOut = (0);
      }
    virtual ItfEffect *Instantiate(PluginBase *base) = 0;
};

template<typename T>
struct EffectInfoImpl : public EffectInfoBase
{
  public:
    EffectInfoImpl(wchar_t const *name, wchar_t const *category) : EffectInfoBase(name, category) {}
    EffectInfoImpl &IconName(wchar_t const *icon) { Icon = icon; return *this; }
    EffectInfoImpl &AudioInputs(size32 num) { AIn = num; return *this; }
    EffectInfoImpl &AudioOutputs(size32 num) { AOut = num; return *this; }
    EffectInfoImpl &MidiInputs(size32 num) { MIn = num; return *this; }
    EffectInfoImpl &MidiOutputs(size32 num) { MOut = num; return *this; }
    ItfEffect *Instantiate(PluginBase *plug)
      {
        T *ret = new T(plug, this);
        return ret;
      }
};

/* Derive from EffectBase to get a simple API to effects, where default behaviors take care */
/* of anything you don't care about. */
class EffectBase : public ItfEffect
{
  public:
    SAPS_SUPPORTLIB_API                         EffectBase(PluginBase *plug, EffectInfoBase const *fi);
    SAPS_SUPPORTLIB_API                         ~EffectBase();

    /* Static API introduced by EffectBase */
    SAPS_SUPPORTLIB_API AudioBus               *AddBus(AudioBusConfig const &config, wchar_t const *name, void *userRef = 0);
    SAPS_SUPPORTLIB_API MidiBus                *AddBus(BusConfig const &config, wchar_t const *name, void *userRef = 0);
    SAPS_SUPPORTLIB_API TimeBus                *AddBus(TimeInput const &config, wchar_t const *name, void *userRef = 0);
    SAPS_SUPPORTLIB_API FloatParameter         *AddParameter(FloatRange const &range, wchar_t const *name, wchar_t const *unit, void *userRef = 0);

    /* ItfEffect API */
    SAPS_SUPPORTLIB_API void                    DoEvents(size32 count, ItfEvent **events);
    SAPS_SUPPORTLIB_API void                    Reconfig(ItfConfig *config);
    SAPS_SUPPORTLIB_API size32                  BusCount();
    SAPS_SUPPORTLIB_API ItfBus                 *Bus(size32 index);
    SAPS_SUPPORTLIB_API void                    Commit();
    SAPS_SUPPORTLIB_API void                    Idle(size32 estMs);
    SAPS_SUPPORTLIB_API ItfWindow              *OpenGui();
    SAPS_SUPPORTLIB_API size32                  SamplesTail();
    SAPS_SUPPORTLIB_API void                    EnterRealtime(TimePos const &atPos);
    SAPS_SUPPORTLIB_API void                    Process();
    SAPS_SUPPORTLIB_API void                    LeaveRealtime();
    SAPS_SUPPORTLIB_API void                    Dispose();
    SAPS_SUPPORTLIB_API ItfPlugin              *Plugin();
    SAPS_SUPPORTLIB_API void                   *Extension(wchar_t const *name);

    inline size32                               BlockSize() { return blockSize_; }
    inline Hz                                   SampleRate() { return sampleRate_; }
    inline ItfHost                             *Host() { return host_; }

    PluginBase                                 *plugin_;
    ItfHost                                    *host_;
    EffectInfoBase const                       *info_;
    std::vector<BusImpl *>                      busses_;
    std::vector<ParameterImpl *>                parameters_;
    Hz                                          sampleRate_;
    size32                                      blockSize_;
};

/* Derive from PluginBase to get a simple API to plug-ins, where default behaviors take care */
/* of anything you don't care about. */
class PluginBase : public ItfPlugin, public PluginInfo
{
  public:
    SAPS_SUPPORTLIB_API                         PluginBase(ItfHost *host, wchar_t const *name,
                                                    wchar_t const *authod, wchar_t const*icon,
                                                    unsigned int msvcver = _MSC_FULL_VER);

    template<typename T> void                   AddEffect(T &b)
      {
        AddEffect(new T(b));
      }
    SAPS_SUPPORTLIB_API void                    AddEffect(EffectInfoBase *b);

    SAPS_SUPPORTLIB_API void                    Unload();
    SAPS_SUPPORTLIB_API PluginInfo const       *Info();
    SAPS_SUPPORTLIB_API size32                  CountEffects();
    SAPS_SUPPORTLIB_API EffectInfo const       *Info(size32 index);
    SAPS_SUPPORTLIB_API ItfEffect              *MakeEffect(size32 index);

    SAPS_SUPPORTLIB_API void                   *Extension(wchar_t const *name);

    ItfHost                                    *host_;
    std::vector<EffectInfoBase *>               effects_;
};


#endif  //  SAPS_SupportLib_h
