
#if !defined(NDEBUG)
#define _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
#endif

#include <SAPS/SupportLib.h>
#include <assert.h>
#include <windows.h>


BusImpl::BusImpl(EffectBase *effect, void *userRef)
  : effect_(effect)
  , userRef_(userRef)
{
  effect->host_->AllocateAddress(&address_, this, effect);
}

BusImpl::~BusImpl()
{
  effect_->host_->DeallocateAddress(address_);
}


AudioBus::AudioBus(AudioBusConfig const &config, wchar_t const *name, EffectBase *base, void *userRef)
  : BusImpl(base, userRef)
  , config_(config)
  , name_(name)
{
  pointers_ = 0;
  channelCount_ = 0;
  dummyBuffer_ = NULL;
}

AudioBus::~AudioBus()
{
  delete[] pointers_;
  delete[] dummyBuffer_;
}

float *AudioBus::Channel(size32 index)
{
  /* You can't ask for a channel that you don't have. */
  assert(index < channelCount_);
  if (index >= channelCount_)
  {
    return 0;
  }
  return pointers_[index];
}

ItfBus *AudioBus::Bus()
{
  return this;
}

void AudioBus::Reconfig(ItfConfig *cfg)
{
  delete[] dummyBuffer_;
  dummyBuffer_ = 0;
  delete[] pointers_;
  pointers_ = 0;
  channelCount_ = 0;
}


Address AudioBus::TargetAddress()
{
  return address_;
}

wchar_t const *AudioBus::Name()
{
  return name_.c_str();
}

bool AudioBus::IsInput()
{
  return (config_.Type & 1) != 0;
}

ItfEffect *AudioBus::Effect()
{
  return effect_;
}

bool AudioBus::RequestChannelCount(size32 count)
{
  delete[] pointers_;
  pointers_ = 0;
  bool ret = true;
  if (count < config_.MinChannelCount)
  {
    count = config_.MinChannelCount;
    ret = false;
  }
  if (count > config_.MaxChannelCount)
  {
    count = config_.MaxChannelCount;
    ret = false;
  }
  channelMask_ = (1 << count) - 1;
  channelCount_ = count;
  pointers_ = new float *[count];
  memset(pointers_, 0, sizeof(float *) * count);
  return ret;
}

size32 AudioBus::ChannelCount()
{
  return channelCount_;
}

void AudioBus::SetPointer(size32 channel, float *ptr)
{
  if (channel >= channelCount_)
  {
    ::OutputDebugString(L"Poorly behaved host: SetPointer() on channel that wasn't allocated.\r\n");
    return;
  }
  if (ptr == 0)
  {
    ::OutputDebugString(L"Poorly behaved host: SetPointer() with a NULL pointer.\r\n");
    if (dummyBuffer_ = NULL)
    {
      dummyBuffer_ = new float[effect_->blockSize_];
    }
    pointers_[channel] = dummyBuffer_;
    return;
  }
  pointers_[channel] = ptr;
}

void AudioBus::SetChannelMask(ChannelMask mask)
{
  channelMask_ = mask;
}

ChannelMask AudioBus::BusChannelMask()
{
  ChannelMask ret = ((1 << config_.MaxChannelCount) - 1);
  if (config_.MinChannelCount > 0)
  {
    ret = ret & ~((1 << (config_.MinChannelCount - 1)) - 1);
  }
  return ret;
}

size32 AudioBus::SamplesDelay()
{
  return config_.InputLatencySamples;
}



MidiBus::MidiBus(BusConfig const &config, wchar_t const *name, EffectBase *base, void *userRef)
  : BusImpl(base, userRef)
  , config_(config)
  , name_(name)
{
}

MidiBus::~MidiBus()
{
}

ItfBus *MidiBus::Bus()
{
  return this;
}

void MidiBus::Reconfig(ItfConfig *cfg)
{
}

Address MidiBus::TargetAddress()
{
  return address_;
}

wchar_t const *MidiBus::Name()
{
  return name_.c_str();
}

bool MidiBus::IsInput()
{
  return (config_.Type & 1) != 0;
}

ItfEffect *MidiBus::Effect()
{
  return effect_;
}






TimeBus::TimeBus(BusConfig const &config, wchar_t const *name, EffectBase *base, void *userRef)
  : BusImpl(base, userRef)
  , config_(config)
  , name_(name)
{
}

TimeBus::~TimeBus()
{
}

ItfBus *TimeBus::Bus()
{
  return this;
}

//  obviously bad, but non-crashing values
static Tempo dummyTempo = { sizeof(Tempo), 1, 1, 1, 1, };
static TimePos dummyTime = { sizeof(TimePos), 1, 1, 1, 1, 1, 1, 1, 1 };

void TimeBus::Reconfig(ItfConfig *cfg)
{
  memset(&pointers_, 0, sizeof(pointers_));
  pointers_.beginTempo = &dummyTempo;
  pointers_.endTempo = &dummyTempo;
  pointers_.beginTime = &dummyTime;
  pointers_.endTime = &dummyTime;
}

Address TimeBus::TargetAddress()
{
  return address_;
}

wchar_t const *TimeBus::Name()
{
  return name_.c_str();
}

bool TimeBus::IsInput()
{
  return (config_.Type & 1) != 0;
}

ItfEffect *TimeBus::Effect()
{
  return effect_;
}

void TimeBus::SetPointers(TimeBusConfig const &config)
{
  pointers_ = config;
  if (pointers_.beginTempo == 0)
  {
    ::OutputDebugString(L"Ill behaved host: TimeBusConfig::beginTempo is NULL.\r\n");
    pointers_.beginTempo = &dummyTempo;
  }
  if (pointers_.endTempo == 0)
  {
    ::OutputDebugString(L"Ill behaved host: TimeBusConfig::endTempo is NULL.\r\n");
    pointers_.endTempo = &dummyTempo;
  }
  if (pointers_.beginTime == 0)
  {
    ::OutputDebugString(L"Ill behaved host: TimeBusConfig::beginTime is NULL.\r\n");
    pointers_.beginTime = &dummyTime;
  }
  if (pointers_.endTime == 0)
  {
    ::OutputDebugString(L"Ill behaved host: TimeBusConfig::endTime is NULL.\r\n");
    pointers_.endTime = &dummyTime;
  }
}
